from typing import Optional

from cpl_core.dependency_injection import ServiceProviderABC
from cpl_discord.service import DiscordBotServiceABC
from cpl_query.extension import List

from bot_core.abc.client_utils_abc import ClientUtilsABC
from bot_data.abc.user_joined_server_repository_abc import UserJoinedServerRepositoryABC
from bot_data.model.user import User
from bot_graphql.abc.filter_abc import FilterABC
from bot_graphql.filter.level_filter import LevelFilter
from modules.level.service.level_service import LevelService


class UserFilter(FilterABC):
    def __init__(
        self,
        services: ServiceProviderABC,
        bot: DiscordBotServiceABC,
        client_utils: ClientUtilsABC,
        levels: LevelService,
        user_joined_servers: UserJoinedServerRepositoryABC,
    ):
        FilterABC.__init__(self)

        self._services = services
        self._bot = bot
        self._client_utils = client_utils
        self._levels = levels
        self._user_joined_servers = user_joined_servers

        self._id = None
        self._discord_id = None
        self._name = None
        self._xp = None
        self._minecraft_id = None
        self._ontime = None
        self._level: Optional[LevelFilter] = None
        self._server = None
        self._left_server = None

    def from_dict(self, values: dict):
        if "id" in values:
            self._id = int(values["id"])

        if "discordId" in values:
            self._discord_id = int(values["discordId"])

        if "name" in values:
            self._name = values["name"]

        if "xp" in values:
            self._xp = int(values["xp"])

        if "minecraftId" in values:
            self._minecraft_id = values["minecraftId"]

        if "ontime" in values:
            self._ontime = int(values["ontime"])

        if "level" in values:
            self._level: LevelFilter = self._services.get_service(LevelFilter)
            self._level.from_dict(values["level"])

        if "server" in values:
            from bot_graphql.filter.server_filter import ServerFilter

            self._server: ServerFilter = self._services.get_service(ServerFilter)
            self._server.from_dict(values["server"])

        if "leftServer" in values:
            self._left_server = values["leftServer"]

    def filter(self, query: List[User]) -> List[User]:
        if self._id is not None:
            query = query.where(lambda x: x.id == self._id)

        if self._discord_id is not None:
            query = query.where(
                lambda x: x.discord_id == self._discord_id or str(self._discord_id) in str(x.discord_id)
            )

        if self._name is not None:

            def _get_member(user: User):
                guild = self._bot.get_guild(user.server.discord_id)
                member = guild.get_member(user.discord_id)
                return member is not None and (member.name == self._name or self._name in member.name)

            query = query.where(_get_member)

        if self._xp is not None:
            query = query.where(lambda x: x.xp == self._xp)

        if self._minecraft_id is not None:
            query = query.where(lambda x: x.minecraft_id == self._minecraft_id)

        if self._ontime is not None:
            query = query.where(lambda x: self._client_utils.get_ontime_for_user(x) == self._ontime)

        if self._level is not None:
            levels = self._level.filter(query.select(lambda x: self._levels.get_level(x))).select(lambda x: x.id)
            query = query.where(lambda x: self._levels.get_level(x).id in levels)

        if self._server is not None:
            servers = self._server.filter(query.select(lambda x: x.server)).select(lambda x: x.id)
            query = query.where(lambda x: x.server.id in servers)

        if self._left_server is not None:

            def _has_user_left_server(user: User):
                active_join = self._user_joined_servers.find_active_user_joined_server_by_user_id(user.id)
                return (active_join is None) == self._left_server

            query = query.where(_has_user_left_server)

        return query
