from typing import Optional, List

import discord
from cpl_core.configuration import ConfigurationABC
from cpl_core.database.context import DatabaseContextABC
from cpl_discord.command import DiscordCommandABC
from cpl_discord.service import DiscordBotServiceABC
from cpl_translation import TranslatePipe
from discord import app_commands
from discord.ext import commands
from discord.ext.commands import Context

from bot_core.abc.client_utils_abc import ClientUtilsABC
from bot_core.abc.message_service_abc import MessageServiceABC
from bot_core.helper.command_checks import CommandChecks
from bot_core.logging.command_logger import CommandLogger
from bot_core.pipes.date_time_offset_pipe import DateTimeOffsetPipe
from bot_data.abc.server_repository_abc import ServerRepositoryABC
from bot_data.abc.user_joined_server_repository_abc import UserJoinedServerRepositoryABC
from bot_data.abc.user_joined_voice_channel_repository_abc import (
    UserJoinedVoiceChannelRepositoryABC,
)
from bot_data.abc.user_repository_abc import UserRepositoryABC
from bot_data.abc.user_warnings_repository_abc import UserWarningsRepositoryABC
from modules.base.service.user_warnings_service import UserWarningsService
from modules.level.service.level_service import LevelService
from modules.permission.abc.permission_service_abc import PermissionServiceABC


class UserGroup(DiscordCommandABC):
    def __init__(
        self,
        config: ConfigurationABC,
        logger: CommandLogger,
        message_service: MessageServiceABC,
        bot: DiscordBotServiceABC,
        client_utils: ClientUtilsABC,
        permissions: PermissionServiceABC,
        servers: ServerRepositoryABC,
        db: DatabaseContextABC,
        users: UserRepositoryABC,
        user_joined_servers: UserJoinedServerRepositoryABC,
        user_joined_voice_channel: UserJoinedVoiceChannelRepositoryABC,
        translate: TranslatePipe,
        date: DateTimeOffsetPipe,
        level: LevelService,
        user_warnings: UserWarningsRepositoryABC,
        user_warnings_service: UserWarningsService,
    ):
        DiscordCommandABC.__init__(self)

        self._config = config
        self._logger = logger
        self._message_service = message_service
        self._bot = bot
        self._client_utils = client_utils
        self._permissions = permissions
        self._servers = servers
        self._db = db
        self._users = users
        self._user_joined_servers = user_joined_servers
        self._user_joined_voice_channel = user_joined_voice_channel
        self._t = translate
        self._date = date
        self._level = level
        self._user_warnings = user_warnings
        self._user_warnings_service = user_warnings_service

        self._logger.trace(__name__, f"Loaded command service: {type(self).__name__}")

        self._atr_dict = {
            "xp": self._t.transform("modules.base.user.atr.xp"),
            "ontime": self._t.transform("modules.base.user.atr.ontime"),
        }

        self._atr_list = [(key, self._atr_dict[key]) for key in self._atr_dict]

    async def _handle_atr_calc(
        self,
        ctx: Context,
        atr: str,
        value: int,
        member: discord.Member,
        is_remove=False,
    ):
        if member is None or not isinstance(member, discord.Member):
            member = ctx.author

        server = self._servers.find_server_by_discord_id(ctx.guild.id)
        user = self._users.find_user_by_discord_id_and_server_id(member.id, server.id)

        if atr == "xp":
            if is_remove:
                user.xp -= value
            else:
                user.xp += value
            self._users.update_user(user)
            self._db.save_changes()
            await self._level.check_level(member)

        else:
            await self._message_service.send_interaction_msg(
                ctx.interaction,
                self._t.transform("modules.base.user.error.atr_not_found").format(atr),
            )
            return

        type = "add" if not is_remove else "remove"
        await self._message_service.send_interaction_msg(
            ctx.interaction,
            self._t.transform(f"modules.base.user.{type}.{atr.lower()}").format(atr, value, member.mention),
        )

    @commands.hybrid_group()
    @commands.guild_only()
    async def user(self, ctx: Context):
        pass

    @user.command()
    @commands.guild_only()
    @CommandChecks.check_is_ready()
    async def info(self, ctx: Context, member: Optional[discord.Member] = None, *, wait: int = None):
        self._logger.debug(__name__, f"Received command user-info {ctx}:{member},{wait}")

        is_mod = self._permissions.is_member_moderator(ctx.author)
        if member is not None and not is_mod:
            await self._message_service.send_ctx_msg(ctx, self._t.transform("common.no_permission_message"))
            return

        if member is None or not isinstance(member, discord.Member):
            member = ctx.author

        server = self._servers.find_server_by_discord_id(ctx.guild.id)
        user = self._users.find_user_by_discord_id_and_server_id(member.id, server.id)
        joins = self._user_joined_servers.get_user_joined_servers_by_user_id(user.id)

        embed = discord.Embed(title=member.name, description=member.name, color=int("ef9d0d", 16))

        embed.add_field(name=self._t.transform("modules.base.user.atr.id"), value=member.id)
        embed.add_field(name=self._t.transform("modules.base.user.atr.name"), value=member.name)
        embed.add_field(
            name=self._t.transform("modules.base.user.atr.discord_join"),
            value=self._date.transform(member.created_at),
            inline=False,
        )
        embed.add_field(
            name=self._t.transform("modules.base.user.atr.last_join"),
            value=self._date.transform(member.joined_at),
            inline=False,
        )
        embed.add_field(name=self._t.transform("modules.base.user.atr.xp"), value=str(user.xp))
        embed.add_field(
            name=self._t.transform("modules.base.user.atr.ontime"),
            value=str(self._client_utils.get_ontime_for_user(user)),
        )

        roles = ""
        for role in member.roles:
            roles += f"{role.name}\n"
        embed.add_field(
            name=self._t.transform("modules.base.user.atr.roles"),
            value=roles,
            inline=False,
        )

        if is_mod or member == ctx.author:
            joins_string = ""
            for join in joins:
                joins_string += f"{self._date.transform(join.joined_on)}\n"
            embed.add_field(
                name=self._t.transform("modules.base.user.atr.joins"),
                value=joins_string,
            )

        if is_mod or member == ctx.author:
            lefts_string = ""
            for join in joins:
                if join.leaved_on is None:
                    if lefts_string == "":
                        lefts_string = "/"
                    continue
                lefts_string += f"{self._date.transform(join.leaved_on)}\n"

            embed.add_field(
                name=self._t.transform("modules.base.user.atr.lefts"),
                value=lefts_string,
            )

        if is_mod or member == ctx.author:
            warnings_string = ""
            for warning in self._user_warnings.get_user_warnings_by_user_id(user.id):
                warnings_string += f"{warning.id} - {warning.description}\n"

            embed.add_field(
                name=self._t.transform("modules.base.user.atr.warnings"),
                value=warnings_string,
                inline=False,
            )

        # send to interaction because of sensitive data
        await self._message_service.send_interaction_msg(ctx.interaction, embed, wait_before_delete=wait)
        self._logger.trace(__name__, f"Finished user-info command")

    @user.command()
    @commands.guild_only()
    @CommandChecks.check_is_ready()
    async def get(self, ctx: Context, atr: str, member: discord.Member = None):
        self._logger.debug(__name__, f"Received command user-get {atr} {ctx}:{member}")

        is_mod = self._permissions.is_member_moderator(ctx.author)
        if member is not None and not is_mod:
            await self._message_service.send_ctx_msg(ctx, self._t.transform("common.no_permission_message"))
            return

        if member is None or not isinstance(member, discord.Member):
            member = ctx.author

        server = self._servers.find_server_by_discord_id(ctx.guild.id)
        user = self._users.find_user_by_discord_id_and_server_id(member.id, server.id)

        if atr == "xp":
            value = str(user.xp)

        elif atr == "ontime":
            value = str(self._client_utils.get_ontime_for_user(user))

        else:
            await self._message_service.send_interaction_msg(
                ctx.interaction,
                self._t.transform("modules.base.user.error.atr_not_found").format(atr),
            )
            return

        await self._message_service.send_interaction_msg(
            ctx.interaction,
            self._t.transform(f"modules.base.user.get.{atr.lower()}").format(member.mention, value),
        )

    @get.autocomplete("atr")
    async def get_autocomplete(self, interaction: discord.Interaction, current: str) -> List[app_commands.Choice[str]]:
        return [app_commands.Choice(name=value, value=key) for key, value in self._atr_list]

    @user.command()
    @commands.guild_only()
    @CommandChecks.check_is_ready()
    @CommandChecks.check_is_member_moderator()
    async def set(self, ctx: Context, atr: str, value: int, member: discord.Member = None):
        self._logger.debug(__name__, f"Received command user-set {atr} {ctx}:{member}")

        if member is None or not isinstance(member, discord.Member):
            member = ctx.author

        server = self._servers.find_server_by_discord_id(ctx.guild.id)
        user = self._users.find_user_by_discord_id_and_server_id(member.id, server.id)

        if atr == "xp":
            try:
                user.xp = value
            except TypeError as te:
                self._logger.error(__name__, f"String value couldn't be converted to int", te)
                await self._message_service.send_interaction_msg(
                    ctx.interaction,
                    self._t.transform("modules.base.user.set.error.type_error"),
                )
                return
            else:
                self._users.update_user(user)
                self._db.save_changes()
                await self._level.check_level(member)

        else:
            await self._message_service.send_interaction_msg(
                ctx.interaction,
                self._t.transform("modules.base.user.error.atr_not_found").format(atr),
            )
            return

        await self._message_service.send_interaction_msg(
            ctx.interaction,
            self._t.transform(f"modules.base.user.set.{atr.lower()}").format(member.mention, value),
        )

    @set.autocomplete("atr")
    async def set_autocomplete(self, interaction: discord.Interaction, current: str) -> List[app_commands.Choice[str]]:
        atr_list = [("xp", self._atr_dict["xp"])]
        return [app_commands.Choice(name=value, value=key) for key, value in atr_list]

    @user.command()
    @commands.guild_only()
    @CommandChecks.check_is_ready()
    @CommandChecks.check_is_member_moderator()
    async def add(self, ctx: Context, atr: str, value: int, member: discord.Member = None):
        self._logger.debug(__name__, f"Received command user-add {atr}-={value} {ctx}:{member}")
        await self._handle_atr_calc(ctx, atr, value, member)

    @add.autocomplete("atr")
    async def set_autocomplete(self, interaction: discord.Interaction, current: str) -> List[app_commands.Choice[str]]:
        atr_list = [("xp", self._atr_dict["xp"])]
        return [app_commands.Choice(name=value, value=key) for key, value in atr_list]

    @user.command()
    @commands.guild_only()
    @CommandChecks.check_is_ready()
    @CommandChecks.check_is_member_moderator()
    async def remove(self, ctx: Context, atr: str, value: int, member: discord.Member = None):
        self._logger.debug(__name__, f"Received command user-remove {atr}-={value} {ctx}:{member}")
        await self._handle_atr_calc(ctx, atr, value, member, is_remove=True)

    @remove.autocomplete("atr")
    async def set_autocomplete(self, interaction: discord.Interaction, current: str) -> List[app_commands.Choice[str]]:
        atr_list = [("xp", self._atr_dict["xp"])]
        return [app_commands.Choice(name=value, value=key) for key, value in atr_list]

    @user.command()
    @commands.guild_only()
    @CommandChecks.check_is_ready()
    @CommandChecks.check_is_member_moderator()
    async def reset(self, ctx: Context, atr: str, member: discord.Member = None):
        self._logger.debug(__name__, f"Received command user-reset {atr} {ctx}:{member}")

        if member is None or not isinstance(member, discord.Member):
            member = ctx.author

        server = self._servers.find_server_by_discord_id(ctx.guild.id)
        user = self._users.find_user_by_discord_id_and_server_id(member.id, server.id)

        if atr == "xp":
            user.xp = 0
            self._users.update_user(user)
            self._db.save_changes()
            await self._level.check_level(member)

        elif atr == "ontime":
            self._user_joined_voice_channel.delete_user_joined_voice_channel_by_user_id(user.id)
            self._db.save_changes()

        else:
            await self._message_service.send_interaction_msg(
                ctx.interaction,
                self._t.transform("modules.base.user.error.atr_not_found").format(atr),
            )
            return

        await self._message_service.send_interaction_msg(
            ctx.interaction,
            self._t.transform(f"modules.base.user.reset.{atr.lower()}").format(atr, member.mention),
        )

    @reset.autocomplete("atr")
    async def reset_autocomplete(
        self, interaction: discord.Interaction, current: str
    ) -> List[app_commands.Choice[str]]:
        return [app_commands.Choice(name=value, value=key) for key, value in self._atr_list]

    @user.group()
    @commands.guild_only()
    async def warning(self, ctx: Context):
        pass

    @warning.command()
    @commands.guild_only()
    @CommandChecks.check_is_ready()
    @CommandChecks.check_is_member_moderator()
    async def show(self, ctx: Context, member: discord.Member, wait: int = None):
        self._logger.debug(__name__, f"Received command user warning show {ctx}:{member}")

        server = self._servers.find_server_by_discord_id(ctx.guild.id)
        user = self._users.find_user_by_discord_id_and_server_id(member.id, server.id)

        embed = discord.Embed(
            title=member.name, description=self._t.transform("modules.base.user.atr.warnings"), color=int("ef9d0d", 16)
        )

        warnings_id_string = ""
        for warning in self._user_warnings.get_user_warnings_by_user_id(user.id):
            warnings_id_string += f"{warning.id}\n"

        warnings_description_string = ""
        for warning in self._user_warnings.get_user_warnings_by_user_id(user.id):
            warnings_description_string += f"{warning.description}\n"

        embed.add_field(
            name=self._t.transform("modules.base.warnings.show.id"),
            value=warnings_id_string,
            inline=True,
        )
        embed.add_field(
            name=self._t.transform("modules.base.warnings.show.description"),
            value=warnings_description_string,
            inline=True,
        )
        await self._message_service.send_interaction_msg(ctx.interaction, embed, wait_before_delete=wait)
        self._logger.trace(__name__, f"Finished user warning show command")

    @warning.command()
    @commands.guild_only()
    @CommandChecks.check_is_ready()
    @CommandChecks.check_is_member_moderator()
    async def add(self, ctx: Context, member: discord.Member, description: str):
        self._logger.debug(__name__, f"Received command user warning add {ctx}:{member},{description}")
        try:
            await self._message_service.send_ctx_msg(ctx, self._t.transform("modules.base.warnings.add.success"))
            await self._user_warnings_service.add_warnings(member, description, ctx.author.id)
        except Exception as e:
            self._logger.error(__name__, f"Adding user warning failed", e)
            await self._message_service.send_ctx_msg(ctx, self._t.transform("modules.base.warnings.add.failed"))
        self._logger.trace(__name__, f"Finished user warning add command")

    @warning.command()
    @commands.guild_only()
    @CommandChecks.check_is_ready()
    @CommandChecks.check_is_member_moderator()
    async def remove(self, ctx: Context, warning_id: int):
        self._logger.debug(__name__, f"Received command user warning remove {ctx}:{warning_id}")
        try:
            await self._message_service.send_ctx_msg(ctx, self._t.transform("modules.base.warnings.remove.success"))
            await self._user_warnings_service.remove_warnings(warning_id)
        except Exception as e:
            self._logger.error(__name__, f"Removing user warning failed", e)
            await self._message_service.send_ctx_msg(ctx, self._t.transform("modules.base.warnings.remove.failed"))
        self._logger.trace(__name__, f"Finished user warning remove command")
