import { Component, OnInit } from "@angular/core";
import { FormBuilder, FormControl, FormGroup, Validators } from "@angular/forms";
import { AuthService } from "src/app/services/auth/auth.service";
import { AuthUserDTO } from "src/app/models/auth/auth-user.dto";
import { ActivatedRoute, Router } from "@angular/router";
import { catchError } from "rxjs/operators";
import { ErrorDTO } from "src/app/models/error/error-dto";
import { AuthErrorMessages } from "src/app/models/auth/auth-error-messages.enum";
import { ServiceErrorCode } from "src/app/models/error/service-error-code.enum";
import { AuthUserAtrErrors } from "src/app/models/auth/auth-user-atr-errors";
import { SpinnerService } from "src/app/services/spinner/spinner.service";
import { ThemeService } from "src/app/services/theme/theme.service";
import { throwError } from "rxjs";
import { TranslateService } from "@ngx-translate/core";
import { ConfirmationDialogService } from "../../../../services/confirmation-dialog/confirmation-dialog.service";

@Component({
  selector: "app-login",
  templateUrl: "./login.component.html",
  styleUrls: ["./login.component.scss"]
})
export class LoginComponent implements OnInit {

  loginForm!: FormGroup<{
    email: FormControl<string | null>,
    password: FormControl<string | null>
  }>;
  submitted = false;

  authUserAtrErrors!: AuthUserAtrErrors;
  code!: string;
  state!: string;
  user!: AuthUserDTO;
  oAuthId!: string;

  constructor(
    private authService: AuthService,
    private formBuilder: FormBuilder,
    private router: Router,
    private spinnerService: SpinnerService,
    private themeService: ThemeService,
    private route: ActivatedRoute,
    private confirmDialog: ConfirmationDialogService,
    private translate: TranslateService
  ) {
  }

  ngOnInit(): void {
    this.initLoginForm();
    this.spinnerService.showSpinner();
    this.authService.isUserLoggedInAsync().then(result => {
      if (result) {
        this.router.navigate(["/dashboard"]);
        return;
      }

      this.checkDiscordLogin();
      this.resetStateFlags();
      this.spinnerService.hideSpinner();
    });
  }

  checkDiscordLogin() {
    this.route.queryParams.pipe(catchError(err => {
      this.spinnerService.hideSpinner();
      this.router.navigate(["auth", "login"]).then(() => {
      });
      return throwError(() => err);
    })).subscribe(params => {
        if (!params["code"] || !params["state"]) {
          this.spinnerService.hideSpinner();
          return;
        }

        this.code = params["code"];
        this.state = params["state"];
        this.authService.discordLogin(this.code, this.state).pipe(catchError(err => {
          this.spinnerService.hideSpinner();
          this.router.navigate(["auth", "login"]).then(() => {
          });
          this.state = "";
          this.code = "";
          return throwError(() => err);
        })).subscribe(token => {
          if (token.firstLogin) {
            console.log(1, this.authService.getEMailFromDecodedToken(this.authService.getDecodedToken(token)))
            this.confirmDialog.confirmDialog(
              this.translate.instant(
                "auth.login.message.confirm_email"),
              this.translate.instant(
                "auth.login.message.confirm_email_d",
                { email: this.authService.getEMailFromDecodedToken(this.authService.getDecodedToken(token)) }
              )
            );
          }

          this.authService.saveToken(token);
          this.themeService.loadTheme();
          this.themeService.loadMenu();
          this.spinnerService.hideSpinner();
          this.router.navigate(["/dashboard"]);
        });
      }
    );
  }

  resetStateFlags(): void {
    this.authUserAtrErrors = new AuthUserAtrErrors();
  }

  initLoginForm(): void {
    this.loginForm = this.formBuilder.group({
      email: ["", [Validators.required, Validators.email]],
      password: ["", [Validators.required, Validators.minLength(8)]]
    });
  }

  register(): void {
    this.router.navigate(["/auth/register"]);
  }

  forgotPassword(): void {
    this.router.navigate(["/auth/forgot-password"]);
  }

  login(): void {
    this.submitted = true;
    this.resetStateFlags();

    if (this.loginForm.invalid) {
      return;
    }

    this.spinnerService.showSpinner();
    const user: AuthUserDTO = {
      firstName: "",
      lastName: "",
      email: this.loginForm.value.email ?? null,
      password: this.loginForm.value.password ?? null
    };

    this.authService.login(user)
      .pipe(catchError(error => {
        if (error.error !== null) {
          const err: ErrorDTO = error.error;

          if (err.errorCode === ServiceErrorCode.InvalidData && err.message === AuthErrorMessages.UserIsEmpty) {
            this.authUserAtrErrors.email.required = true;
            this.authUserAtrErrors.password.required = true;
          } else if (err.errorCode === ServiceErrorCode.InvalidUser && err.message === AuthErrorMessages.UserNotFound) {
            this.authUserAtrErrors.email.wrongData = true;
          } else if (err.errorCode === ServiceErrorCode.InvalidUser && err.message === AuthErrorMessages.WrongPassword) {
            this.authUserAtrErrors.password.wrongData = true;
          } else if (err.errorCode === ServiceErrorCode.InvalidUser && err.message === AuthErrorMessages.EMailNotConfirmed) {
            this.authUserAtrErrors.email.notConfirmed = true;
          }
          error.error = null;
        }
        this.spinnerService.hideSpinner();
        throw error;
      }))
      .subscribe(token => {
        this.authService.saveToken(token);
        this.themeService.loadTheme();
        this.themeService.loadMenu();
        this.spinnerService.hideSpinner();
        this.router.navigate(["/dashboard"]);
      });
  }

  discordLogin() {
    this.authService.getDiscordAuthURL().subscribe(url => {
      window.location.href = url.loginUrl;
    });
  }

}
