import { HttpClient, HttpHeaders } from "@angular/common/http";
import { Injectable } from "@angular/core";
import { Router } from "@angular/router";
import { JwtHelperService } from "@auth0/angular-jwt";
import { firstValueFrom, Observable, Subject, Subscription } from "rxjs";
import { catchError } from "rxjs/operators";
import { AdminUpdateUserDTO } from "src/app/models/auth/admin-update-user.dto";
import { AuthRoles } from "src/app/models/auth/auth-roles.enum";
import { AuthUserDTO } from "src/app/models/auth/auth-user.dto";
import { EMailStringDTO } from "src/app/models/auth/email-string.dto";
import { ResetPasswordDTO } from "src/app/models/auth/reset-password.dto";
import { TokenDTO } from "src/app/models/auth/token.dto";
import { UpdateUserDTO } from "src/app/models/auth/update-user.dto";
import { AuthUserSelectCriterion } from "src/app/models/selection/auth-user/auth-user-select-criterion.dto";
import { GetFilteredAuthUsersResultDTO } from "src/app/models/selection/auth-user/get-filtered-auth-users-result.dto";
import { SettingsService } from "../settings/settings.service";
import { SpinnerService } from "../spinner/spinner.service";
import { DiscordAuthURL } from "../../models/auth/discord-auth-url.dto";
import { OAuthDTO } from "../../models/auth/oauth.dto";

@Injectable({
  providedIn: "root"
})
export class AuthService {

  private isLoggedIn!: boolean;
  isLoggedIn$ = new Subject<boolean>();

  constructor(
    private appsettings: SettingsService,
    private http: HttpClient,
    private router: Router,
    private jwtHelper: JwtHelperService,
    private spinnerService: SpinnerService
  ) {
    this.isLoggedIn$.subscribe(value => {
      this.isLoggedIn = value;
    });
  }

  /* data requests */
  getAllUsers(): Observable<Array<AuthUserDTO>> {
    return this.http.get<Array<AuthUserDTO>>(`${this.appsettings.getApiURL()}/api/auth/users`, {
      headers: new HttpHeaders({
        "Content-Type": "application/json"
      })
    });
  }

  getFilteredUsers(selectCriterions: AuthUserSelectCriterion): Observable<GetFilteredAuthUsersResultDTO> {
    return this.http.post<GetFilteredAuthUsersResultDTO>(`${this.appsettings.getApiURL()}/api/auth/users/get/filtered`, selectCriterions, {
      headers: new HttpHeaders({
        "Content-Type": "application/json"
      })
    });
  }

  getUserByEMail(email: string): Observable<AuthUserDTO> {
    return this.http.get<AuthUserDTO>(`${this.appsettings.getApiURL()}/api/auth/users/get/${email}`, {
      headers: new HttpHeaders({
        "Content-Type": "application/json"
      })
    });
  }

  findUserByEMail(email: string): Observable<AuthUserDTO> {
    return this.http.get<AuthUserDTO>(`${this.appsettings.getApiURL()}/api/auth/users/find/${email}`, {
      headers: new HttpHeaders({
        "Content-Type": "application/json"
      })
    });
  }

  /* auth requsts */
  register(user: AuthUserDTO): Observable<unknown> {
    return this.http.post<TokenDTO>(`${this.appsettings.getApiURL()}/api/auth/register`, user, {
      headers: new HttpHeaders({
        "Content-Type": "application/json"
      })
    });
  }

  confirmEMail(id: string): Observable<boolean> {
    return this.http.post<boolean>(`${this.appsettings.getApiURL()}/api/auth/register-by-id/${id}`, {
      headers: new HttpHeaders({
        "Content-Type": "application/json"
      })
    });
  }

  login(user: AuthUserDTO): Observable<TokenDTO> {
    return this.http.post<TokenDTO>(`${this.appsettings.getApiURL()}/api/auth/login`, user, {
      headers: new HttpHeaders({
        "Content-Type": "application/json"
      })
    });
  }

  verifyLogin(): Observable<boolean> {
    return this.http.get<boolean>(`${this.appsettings.getApiURL()}/api/auth/verify-login`, {
      headers: new HttpHeaders({
        "Content-Type": "application/json"
      })
    });
  }

  forgotPassword(email: string): Observable<unknown> {
    return this.http.post(`${this.appsettings.getApiURL()}/api/auth/forgot-password/${email}`, {
      headers: new HttpHeaders({
        "Content-Type": "application/json"
      })
    });
  }

  getEMailFromforgotPasswordId(id: string): Observable<EMailStringDTO> {
    return this.http.post<EMailStringDTO>(`${this.appsettings.getApiURL()}/api/auth/confirm-forgot-password/${id}`, {
      headers: new HttpHeaders({
        "Content-Type": "application/json"
      })
    });
  }

  resetPassword(resetPasswordDTO: ResetPasswordDTO): Observable<unknown> {
    return this.http.post(`${this.appsettings.getApiURL()}/api/auth/reset-password`, resetPasswordDTO, {
      headers: new HttpHeaders({
        "Content-Type": "application/json"
      })
    });
  }

  updateUser(updateUserDTO: UpdateUserDTO): Observable<unknown> {
    return this.http.post(`${this.appsettings.getApiURL()}/api/auth/update-user`, updateUserDTO, {
      headers: new HttpHeaders({
        "Content-Type": "application/json"
      })
    });
  }

  updateUserAsAdmin(updateUserDTO: AdminUpdateUserDTO): Observable<unknown> {
    return this.http.post(`${this.appsettings.getApiURL()}/api/auth/update-user-as-admin`, updateUserDTO, {
      headers: new HttpHeaders({
        "Content-Type": "application/json"
      })
    });
  }

  refresh(token: TokenDTO): Observable<TokenDTO> {
    return this.http.post<TokenDTO>(`${this.appsettings.getApiURL()}/api/auth/refresh`, token, {
      headers: new HttpHeaders({
        "Content-Type": "application/json"
      })
    });
  }

  deleteUserByMail(mail: string) {
    return this.http.post(`${this.appsettings.getApiURL()}/api/auth/delete-user-by-mail/${mail}`, {
      headers: new HttpHeaders({
        "Content-Type": "application/json"
      })
    });
  }

  /* discord auth */
  getDiscordAuthURL() {
    return this.http.get<DiscordAuthURL>(`${this.appsettings.getApiURL()}/api/auth/discord/get-url`, {
      headers: new HttpHeaders({
        "Content-Type": "application/json"
      })
    });
  }

  discordLogin(code: string, state: string): Observable<TokenDTO> {
    return this.http.get<TokenDTO>(`${this.appsettings.getApiURL()}/api/auth/discord/login?code=${code}&state=${state}`, {
      headers: new HttpHeaders({
        "Content-Type": "application/json"
      })
    });
  }

  // /api/auth/discord/register?code=
  discordRegister(oAuthDTO: OAuthDTO) {
    return this.http.post(`${this.appsettings.getApiURL()}/api/auth/discord/register`, oAuthDTO, {
      headers: new HttpHeaders({
        "Content-Type": "application/json"
      })
    });
  }

  // discordGetUser(code: string, state: string) {
  //   return this.http.get<AuthUserDTO>(`${this.appsettings.getApiURL()}/api/auth/discord/get-user?code=${code}&state=${state}`, {
  //     headers: new HttpHeaders({
  //       'Content-Type': 'application/json'
  //     })
  //   });
  // }

  /* utils */
  saveToken(token: TokenDTO): void {
    localStorage.setItem("jwt", token.token);
    localStorage.setItem("rjwt", token.refreshToken);
    if (this.router.url.startsWith("/auth")) {
      this.router.navigate(["/dashboard"]);
    }
  }

  getToken(): TokenDTO {
    return {
      token: localStorage.getItem("jwt") ?? "",
      refreshToken: localStorage.getItem("rjwt") ?? ""
    };
  }

  getDecodedToken(token: TokenDTO | undefined = undefined): { [key: string]: any } | null {
    if (token) {
    return this.jwtHelper.decodeToken(token.token);
    }

    return this.jwtHelper.decodeToken(this.getToken().token);
  }

  logout(): Subscription | null {
    const token = this.getToken();

    if (token && token.token && token.refreshToken) {
      return this.http.post<TokenDTO>(`${this.appsettings.getApiURL()}/api/auth/revoke`, token, {
        headers: new HttpHeaders({
          "Content-Type": "application/json"
        })
      }).pipe(catchError((error: any) => {
        error.error = null;
        this.isLoggedIn$.next(false);
        localStorage.removeItem("rjwt");
        this.router.navigate(["/auth/login"]);
        throw error;
      })).subscribe(() => {
        this.isLoggedIn$.next(false);
        localStorage.removeItem("jwt");
        localStorage.removeItem("rjwt");
        this.router.navigate(["/auth/login"]);
      });
    }
    this.isLoggedIn$.next(false);
    localStorage.removeItem("rjwt");
    this.router.navigate(["/auth/login"]);

    return null;
  }

  async getLoggedInUser(): Promise<AuthUserDTO | null> {
    if (!await this.isUserLoggedInAsync()) {
      return null;
    }
    const token = this.getDecodedToken();
    if (!token) return null;

    try {
      return await firstValueFrom(this.findUserByEMail(token["email"]));
    } catch (error: unknown) {
      return null;
    }
  }

  async isUserLoggedInAsync(): Promise<boolean> {
    const token = this.getToken();

    if (!token || !token.refreshToken) {
      this.isLoggedIn$.next(false);
      return false;
    }

    try {
      const verifiedLogin = await firstValueFrom(this.verifyLogin());
      if (verifiedLogin) {
        this.isLoggedIn$.next(true);
        return true;
      }
    } catch (error: unknown) {
      this.isLoggedIn$.next(false);
      return false;
    }

    if (this.isLoggedIn) {
      this.spinnerService.showSpinner();

      const resfreshedToken = await firstValueFrom(await this.refresh(token));

      this.spinnerService.hideSpinner();
      if (resfreshedToken) {
        this.saveToken(resfreshedToken);
        return true;
      }
    }

    return false;
  }

  async hasUserPermission(role: AuthRoles): Promise<boolean> {
    if (!role || !await this.isUserLoggedInAsync()) {
      return false;
    }
    const token = this.getDecodedToken();
    if (!token) return false;
    return AuthRoles[token["role"]] === AuthRoles[role];
  }

  getEMailFromDecodedToken(token: { [key: string]: any } | null): string | null {
    if (!token) {
      return null;
    }
    return token["email"];
  }
}
